/* lexxer.l - input source for flex to generate a scanner for lexical
 *            analysis of C source programs
 *
 *  Copyright (C) 2013, Marin Ramesa <marin.ramesa@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Build:
 *   $flex -o lexxer.c lexxer.l
 *   $gcc -o lexxer lexxer.c
 */

%{

#include "token_types.h"
#include "token_classes.h"
#include "output.h"

unsigned long int token_count = 0;
%}

%option yylineno
%option noyywrap

SIGN [-+]?
INVALID_SUFFIX [_a-km-tv-zA-KM-TV-Z]+
NON_ZERO [1-9]
DIGIT [0-9]*
HEX_PREFIX [xX]
HEX_NON_ZERO [1-9a-fA-F]
HEX_DIGIT [0-9a-fA-F]*
OCTAL_NON_ZERO [1-7]
OCTAL_DIGIT [0-7]*
OCTAL_NOT_DEF [89]+
UNSIGNED_SUFFIX [uU]
LONG_SUFFIX [lL]
FLOAT [fF]
EXP [eE]
ID [_a-zA-Z][_a-zA-Z0-9]*

%%

{SIGN}0 { output(CONSTANT, DECIMAL_INTEGER, "decimal integer constant");}

{SIGN}0{INVALID_SUFFIX} {
                                fprintf(stderr, "line %ld: error: "
                                                "expected decimal integer constant token "
                                                "\"0\", but \"%s\" has an invalid suffix.\n",
                                                yylineno, yytext);
                                return 1;
                        }

{SIGN}{NON_ZERO}{DIGIT} { output(CONSTANT, DECIMAL_INTEGER, "decimal integer constant");}

{SIGN}{NON_ZERO}{DIGIT}{INVALID_SUFFIX} {
                                            fprintf(stderr, "line %ld: error: "
                                                            "expected decimal integer constant token, "
                                                            "but \"%s\" has an invalid suffix.\n",
                                                            yylineno, yytext);
                                            return 1;
                                        }

{SIGN}0{HEX_PREFIX}{HEX_NON_ZERO}{HEX_DIGIT} { output(CONSTANT, HEXADECIMAL_INTEGER, "hexadecimal integer constant");}

{SIGN}0{HEX_PREFIX}{HEX_NON_ZERO}{HEX_DIGIT}{INVALID_SUFFIX} {
                                                                fprintf(stderr, "line %ld: error: "
                                                                                "expected hexadecimal integer constant token, "
                                                                                "but \"%s\" has an invalid suffix.\n",
                                                                                yylineno, yytext);
                                                                return 1;
                                                             }

{SIGN}0{OCTAL_NON_ZERO}{OCTAL_DIGIT} { output(CONSTANT, OCTAL_INTEGER, "octal integer constant");}

{SIGN}0{OCTAL_NON_ZERO}{OCTAL_DIGIT}{OCTAL_NOT_DEF} {
                                                        fprintf(stderr, "line %ld: error: "
                                                                        "string \"%s\" contains digits not defined "
                                                                        "in an expected octal integer constant token.\n",
                                                                        yylineno, yytext);
                                                        return 1;
                                                    }

{SIGN}0{OCTAL_NON_ZERO}{OCTAL_NOT_DEF}{OCTAL_DIGIT} {
                                                        fprintf(stderr, "line %ld: error: "
                                                                        "string \"%s\" contains digits not defined "
                                                                        "in an expected octal integer constant token.\n",
                                                                        yylineno, yytext);
                                                        return 1;
                                                    }


{SIGN}0{OCTAL_NON_ZERO}{OCTAL_DIGIT}{INVALID_SUFFIX} {
                                                        fprintf(stderr, "line %ld: error: "
                                                                        "expected octal integer constant token, "
                                                                        "but \"%s\" has an invalid suffix.\n",
                                                                        yylineno, yytext);
                                                        return 1;
                                                     }

0{UNSIGNED_SUFFIX} { output(CONSTANT, DECIMAL_UNSIGNED_INTEGER, "decimal unsigned integer constant");}

{NON_ZERO}{DIGIT}{UNSIGNED_SUFFIX} { output(CONSTANT, DECIMAL_UNSIGNED_INTEGER, "decimal unsigned integer constant");}

{SIGN}0{LONG_SUFFIX} { output(CONSTANT, DECIMAL_LONG_INTEGER, "decimal long integer constant");}

{SIGN}{NON_ZERO}{DIGIT}{LONG_SUFFIX} { output(CONSTANT, DECIMAL_LONG_INTEGER, "decimal long integer constant");}

0{UNSIGNED_SUFFIX}{LONG_SUFFIX} { output(CONSTANT, DECIMAL_UNSIGNED_LONG_INTEGER, "decimal unsigned long integer constant");}

{NON_ZERO}{DIGIT}{UNSIGNED_SUFFIX}{LONG_SUFFIX} { output(CONSTANT, DECIMAL_UNSIGNED_LONG_INTEGER, "decimal unsigned long integer constant");}

0{LONG_SUFFIX}{UNSIGNED_SUFFIX} { output(CONSTANT, DECIMAL_UNSIGNED_LONG_INTEGER, "decimal unsigned long integer constant");}

{NON_ZERO}{DIGIT}{LONG_SUFFIX}{UNSIGNED_SUFFIX} { output(CONSTANT, DECIMAL_UNSIGNED_LONG_INTEGER, "decimal unsigned long integer constant");}

{SIGN}0{LONG_SUFFIX}{LONG_SUFFIX} { output(CONSTANT, DECIMAL_LONG_LONG_INTEGER, "decimal long long integer constant");}

{SIGN}{NON_ZERO}{DIGIT}{LONG_SUFFIX}{LONG_SUFFIX} { output(CONSTANT, DECIMAL_LONG_LONG_INTEGER, "decimal long long integer constant");}

0{UNSIGNED_SUFFIX}{LONG_SUFFIX}{LONG_SUFFIX} { output(CONSTANT, DECIMAL_UNSIGNED_LONG_LONG_INTEGER, "decimal unsigned long long integer constant");}

{NON_ZERO}{DIGIT}{UNSIGNED_SUFFIX}{LONG_SUFFIX}{LONG_SUFFIX} { output(CONSTANT, DECIMAL_UNSIGNED_LONG_LONG_INTEGER, "decimal unsigned long long integer constant");}

0{LONG_SUFFIX}{LONG_SUFFIX}{UNSIGNED_SUFFIX} { output(CONSTANT, DECIMAL_UNSIGNED_LONG_LONG_INTEGER, "decimal unsigned long long integer constant");}

{NON_ZERO}{DIGIT}{LONG_SUFFIX}{LONG_SUFFIX}{UNSIGNED_SUFFIX} { output(CONSTANT, DECIMAL_UNSIGNED_LONG_LONG_INTEGER, "decimal unsigned long long integer constant");}

"..." { output(MISC, ELLIPSIS, "ellipsis");}

{SIGN}{DIGIT}"."{DIGIT}{EXP}?[0-9]+ { output(CONSTANT, DOUBLE_NUM, "double constant");}

{SIGN}{DIGIT}"."{DIGIT}{EXP} {
                                fprintf(stderr, "line %ld: error: "
                                                "exponent has no digits.\n",
                                                yylineno);
                                return 1;
                             }

{SIGN}{DIGIT}"."{DIGIT}{EXP}?[0-9]+{FLOAT} { output(CONSTANT, FLOAT_NUM, "float constant");}

{SIGN}{DIGIT}"."{DIGIT}{EXP}{FLOAT} {
                                        fprintf(stderr, "line %ld: error: "
                                                        "exponent has no digits.\n",
                                                        yylineno);
                                        return 1;
                                    }

{SIGN}{DIGIT}"."{DIGIT}{EXP}?[0-9]+{LONG_SUFFIX} { output(CONSTANT, LONG_DOUBLE, "long double constant");}

{SIGN}{DIGIT}"."{DIGIT}{EXP}{LONG_SUFFIX} {
                                                fprintf(stderr, "line %ld: error: "
                                                                "exponent has no digits.\n",
                                                                yylineno);
                                                return 1;
                                          }

"auto" { output(KEYWORD, AUTO, "keyword");}

"break" { output(KEYWORD, BREAK, "keyword");}

"case" { output(KEYWORD, CASE, "keyword");}

"char" { output(KEYWORD, CHAR, "keyword");}

"const" { output(KEYWORD, CONST, "keyword");}

"continue" { output(KEYWORD, CONTINUE, "keyword");}

"default" { output(KEYWORD, DEFAULT, "keyword");}

"do" { output(KEYWORD, DO, "keyword");}

"double" { output(KEYWORD, DOUBLE, "keyword");}

"else" { output(KEYWORD, ELSE, "keyword");}

"enum" { output(KEYWORD, ENUM, "keyword");}

"extern" { output(KEYWORD, EXTERN, "keyword");}

"float" { output(KEYWORD, FLOAT, "keyword");}

"for" { output(KEYWORD, FOR, "keyword");}

"goto" { output(KEYWORD, GOTO, "keyword");}

"if" { output(KEYWORD, IF, "keyword");}

"int" { output(KEYWORD, INT, "keyword");}

"long" { output(KEYWORD, LONG, "keyword");}

"register" { output(KEYWORD, REGISTER, "keyword");}

"return" { output(KEYWORD, RETURN, "keyword");}

"short" { output(KEYWORD, SHORT, "keyword");}

"signed" { output(KEYWORD, SIGNED, "keyword");}

"sizeof" { output(KEYWORD, SIZEOF, "keyword");}

"static" { output(KEYWORD, STATIC, "keyword");}

"struct" { output(KEYWORD, STRUCT, "keyword");}

"switch" { output(KEYWORD, SWITCH, "keyword");}

"typedef" { output(KEYWORD, TYPEDEF, "keyword");}

"union" { output(KEYWORD, UNION, "keyword");}

"unsigned" { output(KEYWORD, UNSIGNED, "keyword");}

"void" { output(KEYWORD, VOID, "keyword");}

"volatile" { output(KEYWORD, VOLATILE, "keyword");}

"while" { output(KEYWORD, WHILE, "keyword");}

"inline" { output(KEYWORD, INLINE, "keyword");}

"_Bool" { output(KEYWORD, _BOOL, "keyword");}

"_Complex" { output(KEYWORD, _COMPLEX, "keyword");}

"_Imaginary" { output(KEYWORD, _IMAGINARY, "keyword");}

{ID} { output(IDENTIFIER, 0, "identifier");}

"+=" { output(OPERATOR, ASSIGNMENT, "assignment operator");}

"-=" { output(OPERATOR, ASSIGNMENT, "assignment operator");}

"*=" { output(OPERATOR, ASSIGNMENT, "assignment operator");}

"/=" { output(OPERATOR, ASSIGNMENT, "assignment operator");}

"%=" { output(OPERATOR, ASSIGNMENT, "assignment operator");}

"<<=" { output(OPERATOR, ASSIGNMENT, "assignment operator");}

">>=" { output(OPERATOR, ASSIGNMENT, "assignment operator");}

"&=" { output(OPERATOR, ASSIGNMENT, "assignment operator");}

"^=" { output(OPERATOR, ASSIGNMENT, "assignment operator");}

"|=" { output(OPERATOR, ASSIGNMENT, "assignment operator");}

"=" { output(OPERATOR, ASSIGNMENT, "assignment operator");}

"++" { output(OPERATOR, INCREMENT, "increment operator");}

"--" { output(OPERATOR, DECREMENT, "decrement operator");}

"+" { output(OPERATOR, ARITHMETIC, "arithmetic operator");}

"-" { output(OPERATOR, ARITHMETIC, "arithmetic operator");}

"*" { output(OPERATOR, ARITHMETIC, "arithmetic operator");}

"/" { output(OPERATOR, ARITHMETIC, "arithmetic operator");}

"%" { output(OPERATOR, ARITHMETIC, "arithmetic operator");}

"==" { output(OPERATOR, COMPARISON, "comparison operator");}

"!=" { output(OPERATOR, COMPARISON, "comparison operator");}

"<" { output(OPERATOR, COMPARISON, "comparison operator");}

"<=" { output(OPERATOR, COMPARISON, "comparison operator");}

">" { output(OPERATOR, COMPARISON, "comparison operator");}

">=" { output(OPERATOR, COMPARISON, "comparison operator");}

"&&" { output(OPERATOR, LOGICAL, "logical operator");}

"||" { output(OPERATOR, LOGICAL, "logical operator");}

"!" { output(OPERATOR, LOGICAL, "logical operator");}

"<<" { output(OPERATOR, BIT_SHIFT, "bit shift operator");}

">>" { output(OPERATOR, BIT_SHIFT, "bit shift operator");}

"&" { output(OPERATOR, BITWISE_LOGICAL, "bitwise logical operator");}

"|" { output(OPERATOR, BITWISE_LOGICAL, "bitwise logical operator");}

"^" { output(OPERATOR, BITWISE_LOGICAL, "bitwise logical operator");}

"~" { output(OPERATOR, BITWISE_LOGICAL, "bitwise logical operator");}

"," { output(OPERATOR, COMMA, "comma operator");}

"." { output(OPERATOR, MEMBER_ACCESS, "member access operator");}

"->" { output(OPERATOR, MEMBER_ACCESS, "member access operator");}

"?" { output(OPERATOR, CONDITIONAL, "conditional operator");}

":" { output(OPERATOR, CONDITIONAL, "conditional operator");}

"\\" { output(OPERATOR, CONCATENATION, "string concatenation operator");}

"(" { output(SEPARATOR, 0, "separator");}

")" { output(SEPARATOR, 0, "separator");}

"[" { output(SEPARATOR, 0, "separator");}

"]" { output(SEPARATOR, 0, "separator");}

"{" { output(SEPARATOR, 0, "separator");}

"}" { output(SEPARATOR, 0, "separator");}

";" { output(SEPARATOR, 0, "separator");}

L?'(\\.|[^\\'])+' { output(CONSTANT, CHARACTER, "character constant");}

L?'(\\.|[^\\'])+$ {
                        fprintf(stderr, "line %ld: error: "
                                        "unterminated character "
                                        "constant.\n",
                                         yylineno);
                        return 1;
                  }

\"[^\"\n]*\" { output(CONSTANT, STRING, "string constant");}

\"[^\"\n]*$ {
                fprintf(stderr, "line %ld: error: "
                                "unterminated string constant.\n",
                                yylineno);
                return 1;
            }

"/*"        {
                register int c;

                for ( ; ; ) {
                    while ((c = input()) != '*' && c != EOF)
                        ;    /* eat up text of comment */

                    if (c == '*') {
                        while ((c = input()) == '*')
                            ;
                        if (c == '/')
                            break;    /* found the end */
                    }

                    if (c == EOF) {
                        printf("L%ld: error: reached end of file in a comment.\n", yylineno);
                        return 1;
                    }
                }
            }

"//"    {
            register c;
            while ((c = input()) != '\n')
                ;
            if (c == '\n')
                break;
        }

[ \t\n]+

. { printf("L%ld: error: unrecognized character \"%s\".\n", yylineno, yytext); return 1;}

%%

void print_id()
{
    printf("id %ld:", token_count);
}

void print_line()
{
    printf(" line %ld:", yylineno);
}

void print_class(int class)
{
    printf(" class %d:", class);
}

void print_type(int type)
{
    printf(" type %d:", type);
}

void print_symbol()
{
    printf(" \"%s\":", yytext);
}

void print_description(char s[])
{
    printf(" %s\n", s);
}

void output(int class, int type, char desc[])
{
    token_count++;
    print_id();
    print_line();
    print_class(class);
    print_type(type);
    print_symbol();
    print_description(desc);
}

int main(int argc, char *argv[])
{
    ++argv, --argc;
    if (argc > 0)
        yyin = fopen(argv[0], "r");
    else
        yyin = stdin;
    if (yylex() == 1)
        return 1;
    else
        return 0;
}
